// *********************************************************
//   NDitt Ver.0.01 -- Next-gen Ditt --                    *
//   Copyright (C) Yumitaro 2011,2012 All rights reserved. *
// *********************************************************


#include <windows.h>
#include <stdlib.h>
#include <mmsystem.h>

#include "fdrawcmd.h"
#include "fdimg.h"

#define	SECTOR_FILL			0xff	// fill byte for formatted sectors

#define	MAX_SECTOR			32
#define	SCANRESULT			(sizeof(FD_TIMED_SCAN_RESULT)+sizeof(FD_TIMED_ID_HEADER)*MAX_SECTOR)

#define	OPT_PROTECT			0x01
#define	OPT_VERBOSE			0x02
#define	OPT_OVERTRACK		0x04
#define	OPT_D5INCH			0x08
#define	OPT_SKIPSD			0x10

#define	PROTECT				(Option&OPT_PROTECT)
#define	VERBOSE				(Option&OPT_VERBOSE)
#define	OVERTRACK			(Option&OPT_OVERTRACK)
#define	D5INCH				(Option&OPT_D5INCH)
#define	SKIPSD				(Option&OPT_SKIPSD)

#define	RSTDELAY			86


enum CmdType { cmdNone=0, cmdRead, cmdWrite, cmdUnformat };

const char *comname[] = { "None", "Read", "Write", "Unformat" };
const char *bpsname[] = { "500Kbps", "300Kbps", "250Kbps", "1Mbps" };


BYTE Option = 0;
int Retry = 10;

BYTE SCyl = 0;			// Start cylinder
BYTE ECyl = 255;		// End cylinder

FD_CMD_RESULT CmdRes;
FD_INTERRUPT_STATUS Ist;


///////////////////////////////////////////////////////////////////////////////
void WriteCon( const char *format, ... )
{
	char str[1024];
	DWORD ret;
	
	va_list args;
	va_start( args, format );
	wvsprintf( str, format, args );
	va_end( args );
	
	WriteFile( GetStdHandle( STD_OUTPUT_HANDLE ), str, lstrlen( str ), &ret, NULL );
}


char *LastError( void )
{
	static char str[1024];
	FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(),
					MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ), str, sizeof(str), NULL );
	return GetLastError() ? str : (char *)"no error";
}


///////////////////////////////////////////////////////////////////////////////
// Raw command
///////////////////////////////////////////////////////////////////////////////
BOOL SendRawCommand( HANDLE h, DWORD command, LPVOID ibuf, DWORD isize, LPVOID obuf, DWORD osize )
{
	int rt = 0;
	DWORD ret;
	
	while( rt++ < Retry ){
		if( DeviceIoControl( h, command, ibuf, isize, obuf, osize, &ret, NULL ) )
			return TRUE;
	}
	return FALSE;
}

BOOL CmdRecalibrate( HANDLE h )
{
	return SendRawCommand( h, IOCTL_FDCMD_RECALIBRATE, NULL, 0, &Ist, sizeof(Ist) );
}

BOOL CmdSeek( HANDLE h, BYTE cyl, BYTE head )
{
	FD_SEEK_PARAMS sep = { cyl, head };
	return SendRawCommand( h, IOCTL_FDCMD_SEEK, &sep, sizeof(sep), &Ist, sizeof(Ist) );
}

BOOL CmdReadID( HANDLE h, BYTE head, BYTE flag=FD_OPTION_MFM )
{
	FD_READ_ID_PARAMS rip = { flag, head };
	return SendRawCommand( h, IOCTL_FDCMD_READ_ID, &rip, sizeof(rip), &CmdRes, sizeof(CmdRes) );
}

BOOL CmdReadData( HANDLE h, BYTE cyl, BYTE head, BYTE start, BYTE count, BYTE size, BYTE *pv, BYTE flag=FD_OPTION_MFM )
{
	FD_READ_WRITE_PARAMS rwp = { flag, head, cyl, head, start, size > 8 ? 8 : size, start+count, 7<<size, 0xff };
	return SendRawCommand( h, IOCTL_FDCMD_READ_DATA, &rwp, sizeof(rwp), pv, count*(128<<rwp.size) );
}

BOOL CmdReadDeletedData( HANDLE h, BYTE cyl, BYTE head, BYTE start, BYTE count, BYTE size, BYTE *pv, BYTE flag=FD_OPTION_MFM )
{
	FD_READ_WRITE_PARAMS rwp = { flag, head, cyl, head, start, size > 8 ? 8 : size, start+count, 7<<size, 0xff };
	return SendRawCommand( h, IOCTL_FDCMD_READ_DELETED_DATA, &rwp, sizeof(rwp), pv, count*(128<<rwp.size) );
}

BOOL CmdWriteData( HANDLE h, BYTE cyl, BYTE head, BYTE start, BYTE count, BYTE size, BYTE *pv, BYTE flag=FD_OPTION_MFM )
{
	FD_READ_WRITE_PARAMS rwp = { flag, head, cyl, head, start, size > 8 ? 8 : size, start+count, 7<<size, 0xff };
	return SendRawCommand( h, IOCTL_FDCMD_WRITE_DATA, &rwp, sizeof(rwp), pv, count*(128<<rwp.size) );
}

BOOL CmdWriteDeletedData( HANDLE h, BYTE cyl, BYTE head, BYTE start, BYTE count, BYTE size, BYTE *pv, BYTE flag=FD_OPTION_MFM )
{
	FD_READ_WRITE_PARAMS rwp = { flag, head, cyl, head, start, size > 8 ? 8 : size, start+count, 7<<size, 0xff };
	return SendRawCommand( h, IOCTL_FDCMD_WRITE_DELETED_DATA, &rwp, sizeof(rwp), pv, count*(128<<rwp.size) );
}

BOOL CmdVerify( HANDLE h, BYTE cyl, BYTE head, BYTE start, BYTE end, BYTE size, BYTE flag=FD_OPTION_MFM )
{
	FD_READ_WRITE_PARAMS rwp = { flag, head, cyl, head, start, size > 8 ? 8 : size, end, 7<<size, 0xff };
	return SendRawCommand( h, IOCTL_FDCMD_VERIFY, &rwp, sizeof(rwp), NULL, 0 );
}

BOOL CmdFormat( HANDLE h, PFD_FORMAT_PARAMS pfp, ULONG size )
{
	return SendRawCommand( h, IOCTL_FDCMD_FORMAT_TRACK, pfp, size, NULL, 0 );
}

BOOL CmdSpecify( HANDLE h, BYTE srt, BYTE hut, BYTE hlt )
{
	FD_SPECIFY_PARAMS sp = { (srt<<4)|(hut&0x0f), (hlt<<1)|1 };
	return SendRawCommand( h, IOCTL_FDCMD_SPECIFY, &sp, sizeof(sp), &CmdRes, sizeof(CmdRes) );
}


BOOL FDCReset( HANDLE h )
{
	return SendRawCommand( h, IOCTL_FD_RESET, NULL, 0, NULL, 0 );
}

BOOL SetDataRate( HANDLE h, BYTE rate )
{
	return SendRawCommand( h, IOCTL_FD_SET_DATA_RATE, &rate, sizeof(rate), NULL, 0 );
}

BOOL ScanTrack( HANDLE h, BYTE head, PFD_SCAN_RESULT psr, ULONG size, BYTE flag=FD_OPTION_MFM )
{
	FD_SCAN_PARAMS scp = { flag, head };
	return SendRawCommand( h, IOCTL_FD_SCAN_TRACK, &scp, sizeof(scp), psr, size );
}

BOOL TimedScanTrack( HANDLE h, BYTE head, PFD_TIMED_SCAN_RESULT psr, BYTE flag=FD_OPTION_MFM )
{
	FD_SCAN_PARAMS scp = { flag, head };
	return SendRawCommand( h, IOCTL_FD_TIMED_SCAN_TRACK, &scp, sizeof(scp), psr, SCANRESULT );
}

BOOL GetResult( HANDLE h )
{
	return SendRawCommand( h, IOCTL_FD_GET_RESULT, NULL, 0, &CmdRes, sizeof(CmdRes) );
}

BOOL SetSectorOffset( HANDLE h, BYTE offset=0 )
{
	FD_SECTOR_OFFSET_PARAMS sop = { offset };
	return SendRawCommand( h, IOCTL_FD_SET_SECTOR_OFFSET, &sop, sizeof(sop), NULL, 0 );
}

BOOL SetShortWrite( HANDLE h, DWORD size, DWORD delay=0 )
{
	FD_SHORT_WRITE_PARAMS swp = { size, delay };
	return SendRawCommand( h, IOCTL_FD_SET_SHORT_WRITE, &swp, sizeof(swp), NULL, 0 );
}

BOOL MotorOn( HANDLE h )
{
	return SendRawCommand( h, IOCTL_FD_MOTOR_ON, NULL, 0, NULL, 0 );
}

BOOL MotorOff( HANDLE h )
{
	return SendRawCommand( h, IOCTL_FD_MOTOR_OFF, NULL, 0, NULL, 0 );
}

BOOL SyncIndex( HANDLE h )
{
	return SendRawCommand( h, IOCTL_FD_WAIT_INDEX, NULL, 0, NULL, 0 );
}




///////////////////////////////////////////////////////////////////////////////
HANDLE OpenFDD( int drv, BYTE rate )
{
	char str[32];
	
	wsprintf( str, "\\\\.\\fdraw%u", drv&1 );
	
	HANDLE h = CreateFile( str, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
    return( h != INVALID_HANDLE_VALUE && SetDataRate( h, rate ) ) ? h : NULL;
}


///////////////////////////////////////////////////////////////////////////////
BOOL IsFDD( const char *str, int *pDrvNum )
{
	if( !pDrvNum || !isalpha( str[0] ) || str[1] != ':' || str[2] )
		return FALSE;
	
	*pDrvNum = (str[0] & ~0x20) - 'A';
	
	return TRUE;
}


DWORD GetDriverVersion( void )
{
	DWORD ver = 0;
	HANDLE h = CreateFile( "\\\\.\\fdrawcmd", GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
	if( h != INVALID_HANDLE_VALUE ){
		SendRawCommand( h, IOCTL_FDRAWCMD_GET_VERSION, NULL, 0, &ver, sizeof(ver) );
		CloseHandle( h );
	}
	return ver;
}



///////////////////////////////////////////////////////////////////////////////
BOOL ReadTrack( HANDLE h, BYTE cyl, BYTE head, cFDTRK *track )
{
	BYTE rbuf[SCANRESULT];
	PFD_TIMED_SCAN_RESULT psr = (PFD_TIMED_SCAN_RESULT)rbuf;
	PFD_TIMED_ID_HEADER ph = psr->Header;
	BYTE flag = 0;
	
	psr->count = 0;
	track->Unformat();
	
	// ΏۃgbNXL
	if( !SKIPSD ){
		// Px`FbN
		if( !TimedScanTrack( h, head, psr, flag ) ){
			WriteCon( "<<Inaccessible track>>\n" );
			CmdRecalibrate( h );
			CmdRecalibrate( h );
			return TRUE;
		}
	}
	
	// {x`FbN
	if( !psr->count ){
		if( !TimedScanTrack( h, head, psr, FD_OPTION_MFM ) ){
			WriteCon( "<<Inaccessible track>>\n" );
			CmdRecalibrate( h );
			CmdRecalibrate( h );
			return TRUE;
		}
		flag = FD_OPTION_MFM;
	}
	
	if( psr->count ){
		if( VERBOSE ){
			WriteCon( "Sectors:%2u %3s %urpm\n", psr->count, flag&FD_OPTION_MFM ? "MFM" : "FM ", 60000000/psr->tracktime );
			WriteCon( "[Sec] [C] [H] [R] [N]  [Size]\n" );
		}
		
		// ZN^ɓǍ
		for( int i=1; i < psr->count+1; i++, ph++ ){
			if( VERBOSE ) WriteCon( "  %2u: %3u %3u %3u %3u :%5u", i, ph->cyl, ph->head, ph->sector, ph->size, 128<<ph->size );
			cFDSEC *sec = track->AddSector( ph->cyl, ph->head, ph->sector, ph->size );
			
			if( ph->size > 8 ){
				WriteCon( " Over size\n" );
				break;
			}
			
			if( CmdReadData( h, ph->cyl, ph->head, ph->sector, 1, ph->size, sec->Buffer(), flag ) ){
				GetResult( h );
				sec->SetDeleted( CmdRes.st2&0x40 ? 0x10 : 0 );
				if( VERBOSE ){
					if( CmdRes.st2&0x40 ) WriteCon( "[Deleted]" );
					WriteCon( " -> OK\n" );
				}
			}else{
				WriteCon( " -> %s", LastError() );
				GetResult( h );
				sec->SetStatus( 0xb0 );	// G[R[h͂Ƃ肠łB0H(CRC error)ɂĂ
			}
		}
	}else
		if( VERBOSE ) WriteCon( " <No sector>\n" );
	
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
BOOL WriteTrack( HANDLE h, BYTE cyl, BYTE head, cFDTRK *track )
{
	if( track->Count() ){
		BYTE flag = track->Density() ? 0 : FD_OPTION_MFM;
		
		if( VERBOSE ){
			WriteCon( "Sectors:%2u %s%\n", track->Count(), flag&FD_OPTION_MFM ? "MFM" : "FM " );
			WriteCon( "[Sec] [C] [H] [R] [N]  [Size]\n" );
		}
		
		// ZN^ɏ
		for( int i=1; i < track->Count()+1; i++ ){
			cFDSEC *sec = track->Sector( i );
			if( VERBOSE ) WriteCon( "  %2u: %3u %3u %3u %3u :%5u", i,
									sec->C(), sec->H(), sec->R(), sec->N(), sec->DataSize() );
			
			if( sec->Status() ){
				// Xe[^XُZN^͂Ƃ肠łCRC errorƂď
				if( VERBOSE ) WriteCon( "[CRC error]" );
				SetShortWrite( h, (128<<sec->N()), RSTDELAY );
			}
			
			BOOL ret;
			if( sec->Deleted() ){
				if( VERBOSE ) WriteCon( "[Deleted]" );
				ret = CmdWriteDeletedData( h, sec->C(), sec->H(), sec->R(), 1, sec->N(), sec->Buffer(), flag );
			}else
				ret = CmdWriteData( h, sec->C(), sec->H(), sec->R(), 1, sec->N(), sec->Buffer(), flag );
			
			if( ret ){
				if( VERBOSE ) WriteCon( " -> OK\n" );
			}else
				WriteCon( " -> %s", LastError() );
		}
	}else
		if( VERBOSE ) WriteCon( " <No sector>\n" );
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
BOOL UnformatTrack( HANDLE h, BYTE cyl, BYTE head )
{
	BYTE buf[sizeof(FD_FORMAT_PARAMS)+sizeof(FD_ID_HEADER)];
	
	PFD_FORMAT_PARAMS pfp = (PFD_FORMAT_PARAMS)buf;
	PFD_ID_HEADER ph      = pfp->Header;
	
	pfp->flags   = FD_OPTION_MFM;
	pfp->phead   = head;
	pfp->size    = 7;
	pfp->sectors = 1;
	pfp->gap     = 0;
	pfp->fill    = 0;
	
	ph->cyl    = 0;
	ph->head   = 0;
	ph->sector = 0;
	ph->size   = 0;
	
	if( VERBOSE ){
		WriteCon( "Sectors:%2u Size:%u Gap:%02XH Data:%02XH\n", pfp->sectors, pfp->size, pfp->gap, pfp->fill );
		WriteCon( "[Sec] [C] [H] [R] [N]\n" );
		WriteCon( "  %2u: %3u %3u %3u %3u", 1, ph->cyl, ph->head, ph->sector, ph->size );
	}
	
	if( CmdFormat( h, pfp, sizeof(buf) ) ){
		if( VERBOSE ) WriteCon( " -> OK\n" );
	}else
		WriteCon( " -> %s", LastError() );
	
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
BYTE Byte2N( DWORD size )
{
	BYTE sz = 0;
	while( (size>>sz) > 128 ) sz++;
	return sz;
}

BOOL FormatTrack( HANDLE h, BYTE cyl, BYTE head, cFDTRK *track=NULL )
{
	const BYTE gap3[][9] = {
		{ 0x33, 0x33, 0x50, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74 },	// MFM
		{ 0x1b, 0x2a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a } };	// FM
	
	BYTE buf[sizeof(FD_FORMAT_PARAMS)+sizeof(FD_ID_HEADER)*MAX_SECTOR];
	
	PFD_FORMAT_PARAMS pfp = (PFD_FORMAT_PARAMS)buf;
	PFD_ID_HEADER ph      = pfp->Header;
	
	if( !track || !track->Count() )
		return UnformatTrack( h, cyl, head );
	
	// ZN^݃`FbN
	BYTE secmax = Byte2N( track->Sector( 1 )->DataSize() );
	BYTE secmin = Byte2N( track->Sector( 1 )->DataSize() );
	for( int i=1; i < track->Count()+1; i++ ){
		secmax = max( secmax, Byte2N( track->Sector( i )->DataSize() ) );
		secmin = min( secmin, Byte2N( track->Sector( i )->DataSize() ) );
	}
	
	pfp->flags   = track->Density() ? 0 : FD_OPTION_MFM;
	pfp->phead   = head;
	pfp->size    = secmin;
	pfp->sectors = 0;
	pfp->gap     = gap3[pfp->flags&FD_OPTION_MFM ? 0 : 1][min( secmax, 8 )];
	pfp->fill    = SECTOR_FILL;
	
	for( int i=1; i <= track->Count(); i++ ){
		int nn = track->Sector( i )->N() - secmin + 1;
		for( int j=0; j < nn; j++, ph++ ){
			ph->cyl    = j ? 0 : track->Sector( i )->C();
			ph->head   = j ? 0 : track->Sector( i )->H();
			ph->sector = j ? 0 : track->Sector( i )->R();
			ph->size   = j ? 0 : track->Sector( i )->N();
			pfp->sectors++;
		}
	}
	
	if( VERBOSE ){
		if( secmax == secmin )
			WriteCon( "Sectors:%2u Size:%u", pfp->sectors, pfp->size );
		else
			WriteCon( "Sectors:%2u(%2u) Size:%u-%u[Mixed]", pfp->sectors, track->Count(), pfp->size, secmax );
		WriteCon( " %s Gap:%02XH Data:%02XH\n", pfp->flags&FD_OPTION_MFM ? "MFM" : "FM ", pfp->gap, pfp->fill );
		WriteCon( "[Sec] [C] [H] [R] [N]\n" );
		PFD_ID_HEADER phh = pfp->Header;
		for( int i=1; i < pfp->sectors+1; i++, phh++ )
			WriteCon( "  %2u: %3u %3u %3u %3u\n", i, phh->cyl, phh->head, phh->sector, phh->size );
	}
	
	return CmdFormat( h, pfp, sizeof(FD_FORMAT_PARAMS)+sizeof(PFD_ID_HEADER)*pfp->sectors );
}


///////////////////////////////////////////////////////////////////////////////
BOOL ReadDisk( HANDLE h, cFDimg *fd, BYTE strk, BYTE etrk )
{
	BYTE side = fd->Type()&FD_DOUBLESIDE  ? 2 : 1;
	BYTE step = fd->Type()&FD_DOUBLETRACK ? 1 : 2;
	
	for( BYTE cyl = strk; cyl <= etrk; cyl += step ){
		for( BYTE head = 0; head < side; head++ ){
			if( !CmdSeek( h, cyl, head ) )
				return FALSE;
			WriteCon( "\rRead     Track %2u Side %u ", cyl/step, head );
			if( !ReadTrack( h, cyl, head, fd->Track( cyl, head ) ) )
				return FALSE;
		}
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
BOOL WriteDisk( HANDLE h, cFDimg *fd, BYTE strk, BYTE etrk )
{
	BYTE side = fd->Type()&FD_DOUBLESIDE  ? 2 : 1;
	BYTE step = fd->Type()&FD_DOUBLETRACK ? 1 : 2;
	
	for( BYTE cyl = strk; cyl <= etrk; cyl += step ){
		for( BYTE head = 0; head < side; head++ ){
			if( !CmdSeek( h, cyl, head ) )
				return FALSE;
			if( fd->Track( cyl, head )->Count() ){	// Write
				WriteCon( "\rFormat   Track:%2u Side:%u ", cyl/step, head );
				if( !FormatTrack( h, cyl, head, fd->Track( cyl, head ) ) )
					return FALSE;
				WriteCon( "\rWrite    Track:%2u Side:%u ", cyl/step, head );
				if( !WriteTrack( h, cyl, head, fd->Track( cyl, head ) ) )
					return FALSE;
			}else{		// Unformat
				WriteCon( "\rUnformat Track:%2u Side:%u ", cyl/step, head );
				if( !FormatTrack( h, cyl, head, NULL ) )
					return FALSE;
			}
		}
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
BOOL CheckArg( int argc, char **argv, CmdType *pcom, int *pdrv, BYTE *pfdt,
				char **title, int *retry, BYTE *opt, char **fname )
{
	if( argc < 3 ) return FALSE;
	
	*pcom  = cmdNone;
	*pfdt  = FDUNKNOWN;
	*opt   = 0;
	*fname = NULL;
	
	if     ( !lstrcmpi( argv[1], "r" ) ) *pcom = cmdRead;
	else if( !lstrcmpi( argv[1], "w" ) ) *pcom = cmdWrite;
	else if( !lstrcmpi( argv[1], "u" ) ) *pcom = cmdUnformat;
	else								 return FALSE;
	
	if     ( !lstrcmpi( argv[2], "1D"  ) ) *pfdt = FD1D;
	else if( !lstrcmpi( argv[2], "2D"  ) ) *pfdt = FD2D;
	else if( !lstrcmpi( argv[2], "1DD" ) ) *pfdt = FD1DD;
	else if( !lstrcmpi( argv[2], "2DD" ) ) *pfdt = FD2DD;
	else if( !lstrcmpi( argv[2], "2HD" ) ) *pfdt = FD2HD;
	else								   return FALSE;
	
	for( int i=3; i < argc; i++ ){
		if     ( IsFDD( argv[i], pdrv ) && *pdrv > 1 ) return FALSE;
		else if( argv[i][0] == '-'){
			switch( argv[i][1] ){
			case 'T':
			case 't':
				*title = &argv[i][2];
				break;
			case 'R':
			case 'r':
				*retry = atoi( &argv[i][2] );
				break;
			case 'C':
			case 'c':
				switch( argv[i][2] ){
				case 'S':
				case 's':
					SCyl = atoi( &argv[i][3] );
					break;
				case 'E':
				case 'e':
					ECyl = atoi( &argv[i][3] );
					break;
				default:
					return FALSE;
				}
				break;
			default:
				for( int j=1; j < lstrlen( argv[i] ); j++ )
					switch( argv[i][j] ){
					case 'W':
					case 'w':
						*opt |= OPT_PROTECT;
						break;
					case 'O':
					case 'o':
						*opt |= OPT_OVERTRACK;
						break;
					case '5':
						*opt |= OPT_D5INCH;
						break;
					case 'S':
					case 's':
						*opt |= OPT_SKIPSD;
						break;
					case 'V':
					case 'v':
						*opt |= OPT_VERBOSE;
						break;
					default:
						return FALSE;
					}
			}
		}else
			*fname = argv[i];
	}
	
	if( ECyl == 255 ) ECyl = *pfdt&FD_DOUBLETRACK ? 79 : 39;
	if( !(*pfdt&FD_DOUBLETRACK) ){
		SCyl *= 2;
		ECyl *= 2;
	}
	ECyl = OVERTRACK ? MAX_CYLINDER-1 : ECyl;
	
	if( (*pcom == cmdNone) || (*pfdt == FDUNKNOWN) ||
		(SCyl > ECyl) || (SCyl < 0 ) || (ECyl >= MAX_CYLINDER ) ) return FALSE;
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
void Usage( void )
{
	WriteCon(	"NDitt Ver.0.01 -- Next-gen Ditt --\n"
				"\n"
				"Usage: NDitt <command> <type> [-option] <drive> <image>\n"
				"\n"
				"  command:  r - Create image from floppy disk\n"
				"            w - Write image to floppy disk\n"
				"            u - Unformat\n"
				"  type   :  Media type (1D,2D,1DD,2DD,2HD)\n"
				"  option :  -txxx - Disk title, 16 characters\n"
				"            -rxxx - Retry count, default:10\n"
				"            -csxx - Start cylinder, default:0\n"
				"            -cexx - End   cylinder, default:39 or 79\n"
				"            -w    - Write protect\n"
				"            -o    - Over track, Track:160-163\n"
				"            -5    - 5inch drive\n"
				"            -s    - Skip single density track search\n"
				"            -v    - Verbose\n"
				"  drive  :  FD drive, typically A: or B:\n"
				"  image  :  Disk image to create or write\n"
	);
}


///////////////////////////////////////////////////////////////////////////////
BOOL InitDevice( HANDLE *h, int dnum, BYTE rate )
{
	DWORD ver = GetDriverVersion();
	
	if( !ver ){
		WriteCon( "fdrawcmd.sys is not installed, see: http://simonowen.com/fdrawcmd/\n" );
		return FALSE;
	}
	
	if( HIWORD(ver) != HIWORD(FDRAWCMD_VERSION) ){
		WriteCon( "The installed fdrawcmd.sys is not compatible with this utility.\n" );
		return FALSE;
	}
	
	if( !(*h = OpenFDD( dnum, rate )) ){
		WriteCon( "Failed to open FD drive: %s", LastError() );
		return FALSE;
	}
	
	if( !FDCReset( *h ) ){
		WriteCon( "Failed to initialise controller: %s", LastError() );
		CloseHandle( h );
		return FALSE;
	}
	
	CmdRecalibrate( h );
	CmdRecalibrate( h );
	
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
DWORD GetRPM( HANDLE h )
{
	SyncIndex( h );
	DWORD tt = timeGetTime();
	
	for( int i=0; i < 2; i++ )
		SyncIndex( h );
	
	tt = 60000/( timeGetTime() - tt ) * 4;
	return tt;
}



///////////////////////////////////////////////////////////////////////////////
int main( int argc_, char *argv_[] )
{
	CmdType nCommand = cmdNone;
	BYTE DiskType = FDUNKNOWN;
	int DrvNum = -1;
	HANDLE h = NULL;
	char *fname = NULL;
	char *title = NULL;
	D88DISK fddi;
	BYTE DataRate;
	
	if( !CheckArg( argc_, argv_, &nCommand, &DrvNum, &DiskType, &title, &Retry, &Option, &fname ) ){
		Usage();
		return EXIT_FAILURE;
	}
	
	fddi.SetTitle( title );
	fddi.SetType( DiskType );
	
	DataRate = DiskType&FD_HIGHDENSITY ? 0 : D5INCH ? 1 : 2;
	
	if( !InitDevice( &h, DrvNum, DataRate ) ){
		return EXIT_FAILURE;
	}
	
	WriteCon( "Command    : %s\n", comname[nCommand] );
	WriteCon( "Media type : %c%c%c%s\n", DiskType&FD_DOUBLESIDE  ? '2' : '1',
										 DiskType&FD_HIGHDENSITY ? 'H' : DiskType&FD_DOUBLEDENSITY ? 'D' : 'S',
										 DiskType&FD_DOUBLETRACK ? 'D' : ' ', D5INCH ? "(5inch)" : "" );
	WriteCon( "Drive      : %c:\n", 'A' + DrvNum );
	WriteCon( "Data rate  : %s\n", bpsname[DataRate] );
	
	MotorOn( h );
	
    switch( nCommand ){
		case cmdRead:
			if( ReadDisk( h, &fddi, SCyl, ECyl ) && fddi.WriteToFile( fname ) )
				WriteCon( "\rImage created successfully.\n" );
			else{
				WriteCon( "\n\nRead failed: %s\n", LastError() );
				return EXIT_FAILURE;
			}
			break;
			
		case cmdWrite:
			if( fddi.ReadFromFile( fname ) && WriteDisk( h, &fddi, SCyl, ECyl ) )
				WriteCon( "\rImage written successfully.\n" );
			else{
				WriteCon( "\n\nWrite failed: %s\n", LastError() );
				return EXIT_FAILURE;
			}break;
			
		case cmdUnformat:
			fddi.Unformat();
			if( WriteDisk( h, &fddi, SCyl, ECyl ) )
				WriteCon( "\rUnformat disk successfully.\n" );
			else{
				WriteCon( "\n\nUnformat failed: %s\n", LastError() );
				return EXIT_FAILURE;
			}
			break;
			
		default:
			break;
	}
	
	MotorOff( h );
	
	if( h ) CloseHandle( h );
	return EXIT_SUCCESS;
}
