#include <string.h>

#include "fdimg.h"
#include "d88.h"

// -------------------------------------------------------------
// ZN^NX
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
cFDSEC::cFDSEC( BYTE cc, BYTE hh, BYTE rr, BYTE nn, BYTE del, BYTE sta )
{
	c       = cc;
	h       = hh;
	r       = rr;
	n       = nn;
	deleted = del;
	status  = sta;
	data    = NULL;
	next    = NULL;
	
	Resize( nn );
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
cFDSEC::~cFDSEC( void )
{
	if( data ) delete [] data;
	if( next ) delete next;
}


////////////////////////////////////////////////////////////////
// ep[^擾Eݒ
////////////////////////////////////////////////////////////////
BYTE cFDSEC::C( void ){ return c; }					// C擾
BYTE cFDSEC::H( void ){ return h; }					// H擾
BYTE cFDSEC::R( void ){ return r; }					// R擾
BYTE cFDSEC::N( void ){ return n; }					// N擾
BYTE cFDSEC::Deleted( void ){ return deleted; }		// DELETED MARK擾
void cFDSEC::SetDeleted( BYTE de ){ deleted = de; }	// DELETED MARKݒ
BYTE cFDSEC::Status( void ){ return status; }		// Xe[^X擾
void cFDSEC::SetStatus( BYTE st ){ status = st; }	// Xe[^X擾
WORD cFDSEC::DataSize( void ){ return dsize; }		// f[^TCY擾
BYTE *cFDSEC::Buffer( void ){ return data; }		// f[^ւ̃|C^擾
cFDSEC *cFDSEC::Next( void ){ return next; }		// ̃ZN^ւ̃|C^擾
void cFDSEC::SetNext( cFDSEC *sec ){ next = sec; }	// ̃ZN^ւ̃|C^ݒ


////////////////////////////////////////////////////////////////
// f[^擾
////////////////////////////////////////////////////////////////
BOOL cFDSEC::GetData( BYTE *buf, WORD sz )
{
	if( !buf || !data ) return FALSE;
	return memcpy( buf, data, dsize > sz ? sz : dsize ) ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////
// f[^ݒ
////////////////////////////////////////////////////////////////
BOOL cFDSEC::SetData( BYTE *buf, WORD sz )
{
	if( !buf || !data ) return FALSE;
	return memcpy( data, buf, sz > dsize ? dsize : sz ) ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////
// f[^TCYύX
////////////////////////////////////////////////////////////////
BOOL cFDSEC::Resize( BYTE nn )
{
	if( nn > 8 ) nn = 8;
	
	if( data ) delete [] data;
	
	dsize = 128 << nn;
	if( !(data = new BYTE[dsize]) ){
		dsize = 0;
		return FALSE;
	}
	ZeroMemory( data, dsize );
	return TRUE;
}






// -------------------------------------------------------------
// gbNNX
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
cFDTRK::cFDTRK( void )
{
	sectors = NULL;
	density = 0;
	gap3    = 0x20;
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
cFDTRK::~cFDTRK( void )
{
	Unformat();
}


////////////////////////////////////////////////////////////////
// ep[^擾Eݒ
////////////////////////////////////////////////////////////////
BYTE cFDTRK::Count( void )						// ZN^擾
{
	BYTE cn = 0;
	cFDSEC *psc = sectors;
	while( psc ){
		psc = psc->Next();
		cn++;
	}
	return cn;
}
BYTE cFDTRK::Gap3( void ){ return gap3; }		// GAP3擾
void cFDTRK::SetGap3( BYTE gp ){ gap3 = gp; }	// GAP3ݒ
BYTE cFDTRK::Density( void ){ return density; }		// L^x擾
void cFDTRK::SetDensity( BYTE dn ){ density = dn; }	// L^xݒ


////////////////////////////////////////////////////////////////
// ZN^擾
////////////////////////////////////////////////////////////////
cFDSEC *cFDTRK::Sector( BYTE sc )
{
	cFDSEC *psc = sectors;
	while( --sc ){
		if( psc->Next() )
			psc = psc->Next();
		else
			return NULL;
	}
	return psc;
}

////////////////////////////////////////////////////////////////
// ZN^ǉ
////////////////////////////////////////////////////////////////
cFDSEC *cFDTRK::AddSector( BYTE cc, BYTE hh, BYTE rr, BYTE nn, BYTE sz, BYTE del, BYTE sta )
{
	cFDSEC *ns = new cFDSEC( cc, hh, rr, nn, del, sta );
	if( !ns ) return NULL;
	
	if( !sectors )
		sectors = ns;
	else{
		cFDSEC *ls = sectors;
		while( ls->Next() )
			ls = ls->Next();
		ls->SetNext( ns );
	}
	if( sz != 0xff && sz != nn ) ns->Resize( sz );
	return ns;
}


////////////////////////////////////////////////////////////////
// tH[}bg
////////////////////////////////////////////////////////////////
BOOL cFDTRK::Format( BYTE sc, BYTE cc, BYTE hh, BYTE nn )
{
	Unformat();
	
	for( int i=1; i < sc+1; i++ ){
		if( !AddSector( cc, hh, i, nn ) )
			return FALSE;
	}
	return TRUE;
}


////////////////////////////////////////////////////////////////
// AtH[}bg
////////////////////////////////////////////////////////////////
void cFDTRK::Unformat( void )
{
	if( sectors ){
		delete sectors;
		sectors = NULL;
	}
}




// -------------------------------------------------------------
// tbs[fBXNC[WNX
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
cFDimg::cFDimg( void )
{
	type    = FDUNKNOWN;
	protect = FALSE;
	ZeroMemory( title, sizeof(title) );
	Unformat();
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
cFDimg::~cFDimg( void )
{
}


////////////////////////////////////////////////////////////////
// ep[^擾Eݒ
////////////////////////////////////////////////////////////////
char *cFDimg::Title( void ){ return title; }		// fBXN̖O擾
void cFDimg::SetTitle( const char *t )				// fBXN̖Oݒ
{
	ZeroMemory( title, sizeof(title) );
	if( t ) memcpy( title, t, min( sizeof(title)-1, strlen(t) ) );
}
BYTE cFDimg::Type( void ){ return type; }			// fBXN̎ގ擾
void cFDimg::SetType( BYTE tp ){ type = tp; }		// fBXN̎ސݒ
BOOL cFDimg::Protect( void ){ return protect; }		// CgveNg擾
void cFDimg::SetProtect( BOOL wp ){ protect = wp; }	// CgveNgݒ


////////////////////////////////////////////////////////////////
// ЖʃfBXN?
////////////////////////////////////////////////////////////////
BOOL cFDimg::IsSingleSide( void )
{
	for( int i=1; i < MAX_CYLINDER*2; i += 2 )
		if( track[i].Count() ) return FALSE;
	return TRUE;
}


////////////////////////////////////////////////////////////////
// gbN擾
////////////////////////////////////////////////////////////////
cFDTRK *cFDimg::Track( BYTE cyl, BYTE head )
{
	if( cyl >= MAX_CYLINDER || head > 1 ) return NULL;
	else return &track[cyl*2+head];
}


////////////////////////////////////////////////////////////////
// AtH[}bg
////////////////////////////////////////////////////////////////
void cFDimg::Unformat( void )
{
	for( int i=0; i < MAX_CYLINDER*2; i++ )
		track[i].Unformat();
}


// -------------------------------------------------------------
// D88fBXNC[WNX
// -------------------------------------------------------------
////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
D88DISK::D88DISK( void )
{
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
D88DISK::~D88DISK( void )
{
}


////////////////////////////////////////////////////////////////
// gbNf[^Ǎ
////////////////////////////////////////////////////////////////
BOOL D88DISK::ReadFileTrack( FILE *fp, BYTE trk, cFDTRK *track )
{
	D88SECTINFO sinfo;
	DWORD offset;
	int i=0;
	
	if( !fp || !track || (trk >= MAX_CYLINDER*2) ) return FALSE;
	
	track->Unformat();
	
	// gbNf[^e[uǍ
	fseek( fp, 32 + trk * sizeof(DWORD), SEEK_SET );
	fread( &offset, sizeof(DWORD), 1, fp );
	if( offset ){
		fseek( fp, offset, SEEK_SET );
		do{
			// ZN^Ǎ
			fread( &sinfo, sizeof(D88SECTINFO), 1, fp );
			track->SetDensity( sinfo.density );	// SZN^͂ ƂO
			BYTE sz = 0;
			while( (sinfo.size>>sz) > 128 )	sz++;
			
			cFDSEC *sec = track->AddSector( sinfo.c, sinfo.h, sinfo.r, sinfo.n, sz, sinfo.deleted, sinfo.status );
			
			// ZN^f[^Ǎ
			fread( sec->Buffer(), sizeof(BYTE), sinfo.size, fp );
		}while( ++i < sinfo.sec_nr );
	}
	return TRUE;
}


////////////////////////////////////////////////////////////////
// gbNf[^
////////////////////////////////////////////////////////////////
BOOL D88DISK::WriteFileTrack( FILE *fp, BYTE trk, cFDTRK *track )
{
	D88SECTINFO sinfo;
	DWORD offset=0;
	
	if( !fp || !track || (trk >= MAX_CYLINDER*2) ) return FALSE;
	
	if( track->Count() ){
		fseek( fp, 0, SEEK_END );
		offset = ftell( fp );
		
		for( int i=1; i < track->Count()+1; i++ ){
			cFDSEC *sec = track->Sector( i );
			// ZN^񏑍
			ZeroMemory( &sinfo, sizeof(D88SECTINFO) );
			sinfo.c = sec->C();					// ID  C (V_No Жʂ̏ꍇ=gbNNo)
			sinfo.h = sec->H();					// ID  H (wb_AhX Жʂ̏ꍇ=0)
			sinfo.r = sec->R();					// ID  R (gbÑZN^No)
			sinfo.n = sec->N();					// ID  N (ZN^TCY 0:128 1:256 2:512 3:1024)
			sinfo.sec_nr  = track->Count();		// ̃gbNɑ݂ZN^̐
			sinfo.density = track->Density();	// L^x     0x00:{x   0x40:Px
			sinfo.deleted = sec->Deleted();		// DELETED MARK 0x00:m[} 0x10:DELETED
			sinfo.status  = sec->Status();		// Xe[^X
			sinfo.size	  = sec->DataSize();	// ̃ZN^̃f[^TCY
			if( fwrite( &sinfo, sizeof(D88SECTINFO), 1, fp ) < 1 )
				return FALSE;
			
			// ZN^f[^
			if( fwrite( sec->Buffer(), sizeof(BYTE), sinfo.size, fp ) < sinfo.size )
				return FALSE;
		}
	}
	
	// gbNf[^e[u
	fseek( fp, 32 + trk * sizeof(DWORD), SEEK_SET );
	if( fwrite( &offset, sizeof(DWORD), 1, fp ) < 1 )
		return FALSE;
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// C[Wt@CǍ
//  1D=>2D,1DD=>2DDƂēǍ܂̂Œ(\ȂΕ␳Ă݂)
////////////////////////////////////////////////////////////////
BOOL D88DISK::ReadFromFile( const char *fname )
{
	D88HEADER d88h;
	int step;
	
	FILE *fp = fopen( fname, "rb" );
	if( !fp ) return FALSE;
	
	// wb_Ǎ
	fread( &d88h, sizeof(D88HEADER), 1, fp );
	memcpy( title, d88h.name, 16 );
	title[16] = 0;
	switch( d88h.type ){
	case 0x00:	type = FD2D;		break;
	case 0x10:	type = FD2DD;		break;
	case 0x20:	type = FD2HD;		break;
	case 0x40:	type = FD1DD;		break;	// 1DDitt
	default:	type = FDUNKNOWN;
	}
	protect = d88h.protect;			// CgveNg
	
	// gbNǍ
	step = type&FD_DOUBLETRACK ? 1 : 2;
	for( BYTE cyl = 0; cyl < MAX_CYLINDER; cyl += step )
		for( BYTE head = 0; head < 2; head++ )
			ReadFileTrack( fp, cyl*2/step+head, &track[cyl*2+head] );
	
	fclose( fp );
	
	// ЖʃfBXN̏ꍇ
	if( !(type&FD_HIGHDENSITY) && IsSingleSide() )
		type &= ~FD_DOUBLESIDE;
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// C[Wt@Cɏ
////////////////////////////////////////////////////////////////
BOOL D88DISK::WriteToFile( const char *fname )
{
	D88HEADER d88h;
	int step;
	
	FILE *fp = fopen( fname, "wb" );
	if( !fp ) return FALSE;
	
	// wb_񏑍
	ZeroMemory( &d88h, sizeof(D88HEADER) );
	memcpy( d88h.name, title, 16 );
	d88h.protect = protect ? 0x10 : 0;
	switch( type ){
	case FD1D:	d88h.type = 0x00;	break;
	case FD2D:	d88h.type = 0x00;	break;
	case FD1DD:	d88h.type = 0x10;	break;
	case FD2DD:	d88h.type = 0x10;	break;
	case FD2HD:	d88h.type = 0x20;	break;
	default:	d88h.type = 0x00;
	}
	fwrite( &d88h, sizeof(D88HEADER), 1, fp );
	
	// gbN
	step = type&FD_DOUBLETRACK ? 1 : 2;
	for( BYTE cyl = 0; cyl < MAX_CYLINDER; cyl += step )
		for( BYTE head = 0; head < 2; head++ )
			WriteFileTrack( fp, cyl*2/step+head, &track[cyl*2+head] );
	
	// t@CTCY
	fseek( fp, 0, SEEK_END );
	d88h.size = ftell( fp );
	fseek( fp, 0, SEEK_SET );
	fwrite( &d88h, sizeof(BYTE), 32, fp );
	
	fclose( fp );
	
	return TRUE;
}
